from .sample_data import (
    _get_golf_path,
    _get_xslope_path,
    _get_aeolian_path,
    _get_rcm8_path,
    _get_landsat_path,
    golf,
    xslope,
    aeolian,
    rcm8,
    landsat,
)
from . import cube
